##############################################################################
#
# Copyright (c) 2007 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Common z3c.form test setups

$Id: __init__.py 5280 2025-05-26 14:45:06Z roger.ineichen $
"""
from __future__ import absolute_import
from future import standard_library
from zope.interface import implementer
import six
standard_library.install_aliases()
from builtins import str
from six import string_types
__docformat__ = 'restructuredtext'

import os
import json
import doctest
import urllib.parse

import bs4

import zope.component
import zope.interface
import zope.schema
import zope.location.interfaces
import zope.traversing.testing
import zope.traversing.browser
import zope.traversing.browser.interfaces
import zope.publisher.interfaces.http
from zope.pagetemplate.interfaces import IPageTemplate
from zope.password.interfaces import IPasswordManager
from zope.password.password import PlainTextPasswordManager
from zope.publisher.browser import TestRequest

import z3c.form.interfaces
import z3c.form.widget
import z3c.form.field
import z3c.form.browser
import z3c.macro.tales
import z3c.template.template
from z3c.template.interfaces import IContentTemplate

import p01.publisher.testing
import p01.jsonrpc.testing

import j01.jsonrpc
from j01.jsonrpc import interfaces
from j01.jsonrpc import btn


###############################################################################
#
# jsonrpc testing helper
#
# NOTE: this methods require a p01.testbrowser or zope.testbrowser > 5.0.0
# browser instance providing the wsgi application as browser.testapp
#
###############################################################################

# exceptions
class JSONRPCResponseProssessingError(Exception):
    """JSONRPC repsonse processing error"""

class JSONRPCResponseError(Exception):
    """JSONRPC response error"""


def getJSONRPCResponse(browser):
    try:
        # get json response data from browser instance
        data = browser.json()
        if data is None:
            # raise server response error
            raise JSONRPCResponseError('Missing json repsonse data')
        elif data is not None and data.get('error') is not None:
            # raise server response error
            raise JSONRPCResponseError(data.get('error'))
        else:
            return data
    except:
        raise JSONRPCResponseProssessingError(browser.body)


def getJSONRPCResult(browser):
    data = getJSONRPCResponse(browser)
    return data['result']


def getJSONRPCContent(browser):
    data = getJSONRPCResponse(browser)
    return data['result']['content']


def getJSONRPCNextURL(browser):
    data = getJSONRPCResponse(browser)
    return str(data['result']['nextURL'])


def doJSONRPCResponse(browser):
    """Process jsonrpc server response"""
    data = getJSONRPCResult(browser)
    if isinstance(data, six.string_types):
        return data
    if data.get('nextURL') is not None:
        # process nextURL
        nextURL = data['nextURL']
        browser.open(nextURL)
    elif data.get('nextContentURL') is not None:
        # load content from nextURL
        nextContentURL = data['nextContentURL']
        return j01LoadContent(browser, nextContentURL)
    elif data.get('content') is not None:
        content = data['content']
        if data.get('contentTargetExpression') is not None:
            # inject content into target
            contentTargetExpression = data['contentTargetExpression']
            browser.doInjectContent(contentTargetExpression, content)
            return browser.contents
        else:
            # return plain content if no target expression is given
            return content
    else:
        # return result if no nextURL or content is given. This must be
        # something we don't support in our j01.jsonrpc implementation
        return data


def getBrowserURL(browser):
    """Get the browswer url"""
    url = browser.url
    r = list(urllib.parse.urlparse(url))
    r[0] = ''
    r[1] = ''
    r[4] = ''
    return urllib.parse.urlunparse(r)


def getBrowserHost(browser):
    """Get the browswer url"""
    url = browser.url
    if url is not None:
        return urllib.parse.urlparse(url).netloc


def doJSONRPCRequest(browser, url=None, method=None, params=None, button=None,
    headers=None, data=None, retJSON=False, retBody=False, retResponse=False,
    retResult=False, retContent=False, retNextURL=False, applyResponse=True):
    """WSGI application based jsonrpc call handling

    The browser instance must be a p01.testbrowser Browser instance.

    The jsonrpc server returns the following repsonse:

    version 1.0:
        return {
            'result': result,
            'error': None,
            'id': 'id'
            }

    version 1.1:
        return {
            'version': '1.1',
            'result': result,
            'id': 'id'
            }

    version 2.0:
        return {
            'jsonrpc': '2.0',
            'result': result,
            'id': 'id'
            }

    Or the jsonrpc server returns the following error response:

    version 1.0:
        return {
            'result': None,
            'error': message,
            'id': 'id'
            }

    version 1.1:
        return {
            'version': '1.1',
            'error': message,
            'id': 'id'
            }
    version 2.0:
        return {
            'jsonrpc': '2.0',
            'error': {
                'code': code,
                'message': message,
                'data': data
                },
           'id': 'id'
           }
    """
    if url is None:
        url = getBrowserURL(browser)

    if params is None:
        params = {}
    if button is not None:
        params["j01FormHandlerName"] = button

    if data is None:
        data = {
            "jsonrpc": "2.0",
            "id": "j01.jsonrpc.testing",
            "method": None,
            "params": {},
            }

    # apply/override method
    if method is not None:
        data['method'] = method

    # apply/override params
    for k, v in list(params.items()):
        data['params'][k] = v

    # process request
    body = json.dumps(data)
    __traceback_info__ = (url, headers)
    browser._req_content_type = 'application/json-rpc'
    browser.request('POST', url, data=body.encode('utf8'), headers=headers)

    # prepare result before process response. If you need the result from a
    # processed response, just get them from the browser instance after
    # processing as we do here
    if retBody:
        # return current response.body
        res = browser.body
    elif retJSON:
        # return current response.json
        res = browser.json()
    elif retResponse:
        # return jsonrpc response as json data
        res = getJSONRPCResponse(browser)
    elif retResult:
        # return result data
        res = getJSONRPCResult(browser)
    elif retContent:
        # return content data
        res = getJSONRPCContent(browser)
    elif retNextURL:
        # return nextURL
        res = getJSONRPCNextURL(browser)
    else:
        res = None

    # process response
    # this could force another request if nextURL is given and means,
    # that the current response is not from related to the initial
    # jsonrpc request
    if applyResponse is True:
        # marked for processing
        doJSONRPCResponse(browser)
    elif applyResponse is not None:
        # applyResponse must be a callable if not True or None, just use them
        # for processing
        applyResponse(browser)

    # return the result from the initial jsonrpc response if asked for
    return res


# j01.dialog jsonrpc helper methods
def j01Dialog(browser, url=None, **kwargs):
    method = "j01Dialog"
    return doJSONRPCRequest(browser, url=url, method=method, **kwargs)


def j01DialogContent(browser, url=None, **kwargs):
    method = "j01DialogContent"
    return doJSONRPCRequest(browser, url=url, method=method, **kwargs)


def j01DialogFormProcessor(browser, url=None, **kwargs):
    method = "j01DialogFormProcessor"
    return doJSONRPCRequest(browser, url=url, method=method, **kwargs)


# j01.jsonrpc jsonrpc helper methods
def j01LoadContent(browser, url=None, **kwargs):
    method = "j01LoadContent"
    return doJSONRPCRequest(browser, url=url, method=method, **kwargs)


def j01FormProcessor(browser, url=None, **kwargs):
    """Form processing method

    By default this method will process the given jsonrpc server response as
    the related j01.jsonrpc javacsript library does. If you use different
    jsonrpc response handler methods you probably ned to implement your own
    repsonse processing methods. If so, just skip prost processing by using
    retRaw=True and use the returned browser content or the browser instance
    which contains the respsonse in browser.contents.

    The j01FormProcessor method returns the following response:

    return {'url': pageURL,
            'content': content,
            'contentTargetExpression': targetExpression,
            # animation
            'scrollToExpression': self.context.scrollToExpression,
            'scrollToOffset': self.context.scrollToOffset,
            'scrollToSpeed': self.context.scrollToSpeed,
            # history setup
            'skipState': skipState,
            'stateURL': stateURL,
            'stateTitle': stateTitle,
            'stateCallbackName': stateCallbackName,
            # load next page with location.href redirect
            'nextURL': nextURL,
            # load next content with jsonrpc
            'nextContentURL': nextContentURL,
            # next hash for custom navitation concepts
            'nextHash': self.context.nextHash,
            }
    """
    method = "j01FormProcessor"
    return doJSONRPCRequest(browser, url=url, method=method, **kwargs)


###############################################################################
#
# test setup helper
#
###############################################################################

@implementer(z3c.form.interfaces.IFormLayer)
class FormTestRequest(TestRequest):
    pass


def getPath(filename, package=None):
    if package is None:
        package = __file__
    return os.path.join(os.path.dirname(package), filename)


def setupJSONRPCFormDefaults():
    """setup jsonrpc form defaults"""

    # setup absulutURL adapter
    zope.traversing.testing.setUp()
    zope.component.provideUtility(PlainTextPasswordManager(), IPasswordManager,
        'Plain Text')

    # setup form template
    zope.component.provideAdapter(
        z3c.template.template.TemplateFactory(
            getPath('layout.pt'), 'text/html'),
        (zope.interface.Interface, z3c.form.interfaces.IFormLayer),
        IContentTemplate)

    # setup button widgets
    zope.component.provideAdapter(
        z3c.form.widget.WidgetTemplateFactory(
            getPath('btn_display.pt', j01.jsonrpc.__file__), 'text/html'),
        (None, None, None, None, interfaces.IButtonWidget),
        IPageTemplate, name=z3c.form.interfaces.DISPLAY_MODE)

    zope.component.provideAdapter(
        z3c.form.widget.WidgetTemplateFactory(
            getPath('btn_input.pt', j01.jsonrpc.__file__), 'text/html'),
        (None, None, None, None, interfaces.IButtonWidget),
        IPageTemplate, name=z3c.form.interfaces.INPUT_MODE)

    # setup button action
    zope.component.provideAdapter(btn.ButtonAction,
        provides=z3c.form.interfaces.IButtonAction)

    # setup form adapters for jsonrpc request
    zope.component.provideAdapter(
        z3c.form.button.ButtonAction,
        (p01.jsonrpc.interfaces.IJSONRPCRequest, z3c.form.interfaces.IButton),
        z3c.form.interfaces.IButtonAction)

    zope.component.provideAdapter(z3c.form.field.FieldWidgets,
        (z3c.form.interfaces.IFieldsForm,
         p01.jsonrpc.interfaces.IJSONRPCRequest, zope.interface.Interface))

    zope.component.provideAdapter(
        z3c.form.browser.text.TextFieldWidget,
        adapts=(zope.schema.interfaces.ITextLine,
                p01.jsonrpc.interfaces.IJSONRPCRequest))

    zope.component.provideAdapter(
        z3c.template.template.TemplateFactory(
            getPath('layout.pt'), 'text/html'),
        (zope.interface.Interface, p01.jsonrpc.interfaces.IJSONRPCRequest),
        IContentTemplate)


def browserView(for_, name, factory, providing=zope.interface.Interface):
    zope.component.provideAdapter(factory,
        (for_, zope.publisher.interfaces.http.IHTTPRequest),
        providing, name=name)

def setUpAbsoluteURL():
    browserView(None, "absolute_url", zope.traversing.browser.AbsoluteURL)
    browserView(zope.location.interfaces.IRoot, "absolute_url",
        zope.traversing.browser.SiteAbsoluteURL)
    browserView(None, '', zope.traversing.browser.AbsoluteURL,
        providing=zope.traversing.browser.interfaces.IAbsoluteURL)
    browserView(zope.location.interfaces.IRoot, '',
        zope.traversing.browser.SiteAbsoluteURL,
        providing=zope.traversing.browser.interfaces.IAbsoluteURL)
    browserView(p01.jsonrpc.testing.IJSONRPCTestApplication, '',
        zope.traversing.browser.SiteAbsoluteURL,
        providing=zope.traversing.browser.interfaces.IAbsoluteURL)


###############################################################################
#
# Unittest setup
#
###############################################################################

from zope.pagetemplate.engine import Engine
from zope.pagetemplate.engine import _Engine
from zope.pagetemplate.engine import TrustedEngine
from zope.pagetemplate.engine import _TrustedEngine

def registerType(name, handler):
    Engine.registerType(name, handler)
    TrustedEngine.registerType(name, handler)


def clear():
    Engine.__init__()
    _Engine(Engine)
    TrustedEngine.__init__()
    _TrustedEngine(TrustedEngine)


try:
    from zope.testing.cleanup import addCleanUp
except ImportError:
    pass
else:
    addCleanUp(clear)


def setUp(test):
    registerType('macro', z3c.macro.tales.MacroExpression)


def tearDown(test):
    # ensure that we cleanup everything
    zope.testing.cleanup.cleanUp()


###############################################################################
#
# Test layer
#
###############################################################################

p01.publisher.testing.defineLayer("JSONRPCTestingLayer", "ftesting.zcml",
    allow_teardown=True, pkgName='j01.jsonrpc.testing')


###############################################################################
#
# Doctest setup
#
###############################################################################

def FunctionalDocFileSuite(*paths, **kw):
    # use our custom HTTPCaller and layer
    globs = kw.setdefault('globs', {})
    globs['getRootFolder'] = p01.publisher.testing.getRootFolder
    kw['package'] = doctest._normalize_module(kw.get('package'))
    # _prepare_doctest_keywords(kw)
    p01.publisher.testing.prepareDocTestKeywords(kw)
    suite = doctest.DocFileSuite(*paths, **kw)
    suite.layer = JSONRPCTestingLayer # p01.checker.ignore
    return suite


# ###############################################################################
# #
# # Doctest setup
# # NOTE: this test setup still uses the HTTPCaller given from p01.jsonrpc
# #       We should migrate p01.jsonrpc to a webtest based setup.
# #       See: XMLRPCTestTransport in zope.app.wsgi.testlayer for a sample.
# #
# ###############################################################################

# def _prepare_doctest_keywords(kw):
#     globs = kw.setdefault('globs', {})
#     globs['http'] = p01.jsonrpc.testing.HTTPCaller()
#     globs['getRootFolder'] = functional.getRootFolder
#     globs['sync'] = functional.sync

#     kwsetUp = kw.get('setUp')
#     def setUp(test):
#         functional.FunctionalTestSetup().setUp()
#         if kwsetUp is not None:
#             kwsetUp(test)
#     kw['setUp'] = setUp

#     kwtearDown = kw.get('tearDown')
#     def tearDown(test):
#         if kwtearDown is not None:
#             kwtearDown(test)
#         functional.FunctionalTestSetup().tearDown()
#     kw['tearDown'] = tearDown

#     if 'optionflags' not in kw:
#         old = doctest.set_unittest_reportflags(0)
#         doctest.set_unittest_reportflags(old)
#         kw['optionflags'] = (old
#                              | doctest.ELLIPSIS
#                              | doctest.REPORT_NDIFF
#                              | doctest.NORMALIZE_WHITESPACE)


# def FunctionalDocFileSuite(*paths, **kw):
#     # use our custom HTTPCaller and layer
#     kw['package'] = doctest._normalize_module(kw.get('package'))
#     _prepare_doctest_keywords(kw)
#     suite = doctest.DocFileSuite(*paths, **kw)
#     suite.layer = JSONRPCTestingLayer # p01.checker.ignore
#     return suite



#======
#README
#======

#Let's test the source code with our custom checker:

from __future__ import absolute_import
import p01.checker#>>>
import j01.jsonrpc#>>>
skipFileNames = [#>>>
  'jquery-migrate-1.4.1.js',#...
  'jquery-migrate-1.4.1.min.js',#...
]#...

checker = p01.checker.Checker(p01.checker.CHECKERS)#>>>
checker.check(j01.jsonrpc, skipFileNames=skipFileNames)#>>>
#  -------------------
#  css/j01.loading.css
#  -------------------
#  63: Unknown @ rule: @-ms-keyframes.
#        @-ms-keyframes pulse {
#  73: Unexpected token '}' at line 73, col 1.
#        }



###############################################################################
#
# Copyright (c) 2015 Projekt01 GmbH
# All Rights Reserved.
#
###############################################################################
"""Links
$Id:$
"""
__docformat__ = "reStructuredText"

UNDEFINED = object()

def isTrue(value):
    if value in [True, 'True', 'true', '1', 1, 'push']:
        return True
    else:
        # anything else then: True, 'True', 'true', '1', 'push'
        return False

def isFalse(value):
    if value in [False, 'False', 'false', '0', 0]:
        return True
    else:
        # anything else then: False, 'False', 'false', '0', 0
        return False


###############################################################################
#
# link

LINK = """<a href="%(url)s"%(attrs)s>%(content)s</a>
"""

def getLoadContentLink(request, url=None, content=None, **kws):
    """Returns a clickable <a> tag supporting testing markers

    Note: the rendered dom element is supported by p01.testbrowser testing!
    You can test this with something like:

        link = browser.getLink(text=None, url=None, id=None, index=0)
        link.click()

    """
    # get default attributes
    isPushState = kws.pop('isPushState', UNDEFINED)
    onSuccess = kws.pop('onSuccess', 'j01RenderContent')
    onError = kws.pop('onError', 'j01RenderContent')
    # setup default attributes
    data = {
        'class': 'j01LoadContentLink',
        }
    # setup testing markers
    if request.get('paste.testing'):
        # testing
        data['data-j01-testing-typ'] = 'JSONRPCClick'
        data['data-j01-testing-url'] = url
        data['data-j01-testing-method'] = 'j01LoadContent'
        data['data-j01-testing-success'] = onSuccess
        data['data-j01-testing-error'] = onError
        # we can set the isPushState in j01.jsonrpc.js, this means we don't know
        # if the default is True or False for testing. The isPushStateTesting
        # marker allows to simulate the j01.jsonrpc.js setup without to define
        # the isPushState at all.
        if isPushState is UNDEFINED:
            isPushState = kws.pop('isPushStateTesting', True)
        # add history state (isPushState) marker
        if isTrue(isPushState):
            data['data-j01-history'] = 'push'
        else:
            data['data-j01-history'] = 'false'
    else:
        # not testing, only inlcude history (isPushState) marker if explicit
        # defined as True/False, otherwise the javascript option defined in
        # j01.jsonrpc.js get used
        if isTrue(isPushState):
            # only add marker if disabled
            data['data-j01-history'] = 'push'
        if isFalse(isPushState):
            data['data-j01-history'] = 'false'
    # update and override default data
    data.update(**kws)
    # render link
    aStr = ''
    for k, v in list(data.items()):
        aStr += ' %s="%s"' % (k, v)
    return LINK % {'url': url, 'attrs': aStr, 'content': content}


###############################################################################
#
# clickable dom element

ELEMENT = """<%(tag)s %(attrs)s>%(content)s</%(tag)s>
"""

def getLoadContentClickable(request, tag=None, url=None, content=None, **kws):
    """Returns a clickable <tag> supporting testing markers

    Note: the rendered dom element is supported by p01.testbrowser testing!
    You can test this with something like:

        clickable = browser.getClickable(text=None, url=None, id=None, index=0)
        clickable.click()

    """
    # get default attributes
    dataURLName = kws.pop('dataURLName', 'url')
    isPushState = kws.pop('isPushState', UNDEFINED)
    onSuccess = kws.pop('onSuccess', 'j01RenderContent')
    onError = kws.pop('onError', 'j01RenderContent')
    # setup default attributes
    data = {
        dataURLName: url,
        'class': 'j01LoadContentClickable',
        }
    # setup testing markers
    if request.get('paste.testing'):
        # testing
        data['data-j01-testing-typ'] = 'JSONRPCClick'
        data['data-j01-testing-url'] = url
        data['data-j01-testing-method'] = 'j01LoadContent'
        data['data-j01-testing-success'] = onSuccess
        data['data-j01-testing-error'] = onError
        # we can set the isPushState in j01.jsonrpc.js, this means we don't know
        # if the default is True or False for testing. The isPushStateTesting
        # marker allows to simulate the j01.jsonrpc.js setup without to define
        # the isPushState at all.
        if isPushState is UNDEFINED:
            isPushState = kws.pop('isPushStateTesting', True)
        # add history state (isPushState) marker
        if isTrue(isPushState):
            data['data-j01-history'] = 'push'
        else:
            data['data-j01-history'] = 'false'
    else:
        # not testing, only inlcude history (isPushState) marker if explicit
        # defined as True/False, otherwise the javascript option defined in
        # j01.jsonrpc.js get used
        if isTrue(isPushState):
            # only add marker if disabled
            data['data-j01-history'] = 'push'
        if isFalse(isPushState):
            data['data-j01-history'] = 'false'
    # update and override default data
    data.update(**kws)
    # render link
    aStr = ''
    for k, v in list(data.items()):
        aStr += ' %s="%s"' % (k, v)
    return ELEMENT % {'tag': tag, 'attrs': aStr, 'content': content}



##############################################################################
#
# Copyright (c) 2007 Projekt01 GmbH and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

"""
$Id: __init__.py 6 2006-04-16 01:28:45Z roger.ineichen $
"""
from __future__ import absolute_import
__docformat__ = 'restructuredtext'

from p01.jsonrpc.publisher import MethodPublisher


class J01LoadContent(MethodPublisher):
    """Can load content via JSONRPC.

    Note: This method publisher must be able to traverse the context which is
    a view or a form. Our default JSON-RPC tarverser could be used.
    See: JSONRPCTraversablePage

    Note: This method is registered with zope.Public permission and will run
    into Unauthorized if the principal does not have the permission to traverse
    the view.

    This method supports the browser history api based on the j01.history.js
    concept.

    Note: it's important that you only provide a stateURL for pages which are
    available as standalone page and can get loaded with a browser request.
    Use the cbStateURL if your page is only loadable with a jsonrpc request.

    """

    def j01LoadContent(self):
        self.context.update()
        if self.context.nextURL is None and self.context.nextContentURL is None:
            content = self.context.render()
        else:
            # skip rendering if a next url is given
            content = None
        # setup response data after render
        data = {
            # url is not used by default callback concept
            'url': self.context.pageURL,
            # redirect url
            'nextURL': self.context.nextURL,
            # load content url
            'nextContentURL': self.context.nextContentURL,
            # content for rendering
            'content': content,
            # jquery content target expression
            'contentTargetExpression': self.context.contentTargetExpression,
            # callback rendering concept, used for invoke javascript rendering
            # if the default content and contentTargetExpression is not enough
            'contentRenderMethodName': getattr(self.context,
                'contentRenderMethodName', None),
            }
        # setup layout helper css classes
        if self.context.cssBodyClass is not None:
            data['cssBodyClass'] = self.context.cssBodyClass
        if self.context.cssHTMLClass is not None:
            data['cssHTMLClass'] = self.context.cssHTMLClass
        if self.context.cssBodyId is not None:
            data['cssBodyId'] = self.context.cssBodyId
        if self.context.cssHTMLId is not None:
            data['cssHTMLId'] = self.context.cssHTMLId
        # setup optional animation
        if self.context.scrollToExpression is not None:
            data['scrollToExpression'] = self.context.scrollToExpression
            data['scrollToOffset'] = self.context.scrollToOffset
            data['scrollToSpeed'] = self.context.scrollToSpeed
        # setup optinal navigation hash useable for for custom navigation
        # concepts (not used in j01.jsonrpc.js but usable for mobile concepts)
        if self.context.nextHash is not None:
            data['nextHash'] = self.context.nextHash
        # setup history state data
        if not self.context.skipState:
            data['state'] = {
                'url': self.context.stateURL,
                'title': self.context.stateTitle,
                'cbURL': self.context.cbStateURL,
                'method': self.context.cbStateMethod,
                'params': self.context.cbStateParams,
                'onSuccess': self.context.cbStateSuccess,
                'onError': self.context.cbStateError,
                'onTimeout': self.context.cbStateTimeout,
                }
        return data


class J01FormProcessor(MethodPublisher):
    """Can process a form button handler via JSONRPC.

    Note: This method publisher must be able to traverse the context which is
    a view or a form. Our default JSON-RPC tarverser could be used.
    See: JSONRPCTraversablePage

    Note: This method is registered with zope.Public permission and will run
    into Unauthorized if the principal does not have the permission to traverse
    the view.
    """

    def j01FormProcessor(self):
        self.context.update()
        if self.context.nextURL is None and self.context.nextContentURL is None:
            content = self.context.render()
        else:
            # skip rendering if a next url is given
            content = None
        # setup response data after render
        data = {
            'url': self.context.pageURL,
            'nextURL': self.context.nextURL,
            'nextContentURL': self.context.nextContentURL,
            'content': content,
            'contentTargetExpression': self.context.contentTargetExpression,
            # callback rendering concept, used for invoke javascript rendering
            # if the default content and contentTargetExpression is not enough
            'contentRenderMethodName': getattr(self.context,
                'contentRenderMethodName', None),
            }
        # setup layout helper css classes
        if self.context.cssBodyClass is not None:
            data['cssBodyClass'] = self.context.cssBodyClass
        if self.context.cssHTMLClass is not None:
            data['cssHTMLClass'] = self.context.cssHTMLClass
        if self.context.cssBodyId is not None:
            data['cssBodyId'] = self.context.cssBodyId
        if self.context.cssHTMLId is not None:
            data['cssHTMLId'] = self.context.cssHTMLId
        # setup optinal animation
        if self.context.scrollToExpression is not None:
            data['scrollToExpression'] = self.context.scrollToExpression
            data['scrollToOffset'] = self.context.scrollToOffset
            data['scrollToSpeed'] = self.context.scrollToSpeed
        # setup optinal navigation hash useable for for custom navigation
        # concepts (not used in j01.jsonrpc.js but usable for mobile concepts)
        if self.context.nextHash is not None:
            data['nextHash'] = self.context.nextHash
        # setup history state data
        if not self.context.skipState:
            data['state'] = {
                'url': self.context.stateURL,
                'title': self.context.stateTitle,
                'cbURL': self.context.cbStateURL,
                'method': self.context.cbStateMethod,
                'params': self.context.cbStateParams,
                'onSuccess': self.context.cbStateSuccess,
                'onError': self.context.cbStateError,
                'onTimeout': self.context.cbStateTimeout,
                }
        return data



##############################################################################
#
# Copyright (c) 2015 Projekt01 GmbH and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
$Id:$
"""
from __future__ import absolute_import
from builtins import object
__docformat__ = "reStructuredText"


class LayoutHelperMixin(object):
    """Layout helper mixin class

    The following attributes can get used for render css class and id for html
    and body elements. This allows you to control css rendering or javascript
    includes.

    You can simply implement your own logik for apply the right ids or css
    classes in your pages within this methods. All you have to do is to use
    them in your layout with something like:

        <!DOCTYPE html>
        <html i18n:domain="demo" lang="de"
              tal:attributes="lang request/lang|nothing;
                              id view/cssHTMLId|nothing;
                              class view/cssHTMLClass|nothing;">
        <head>...</head>
        <body tal:attributes="id view/cssBodyId|nothing;
                              class view/cssBodyClass|nothing;">
        ...
        </body>
        </html>

    and apply the classes and ids like:

        cssHTMLId = 'layout2'
        cssHTMLClass = 'more another'
        cssBodyId = 'body'
        cssBodyClass = 'option1 feature2'

    """

    cssBodyId = None
    cssBodyClass = None
    cssHTMLId = None
    cssHTMLClass = None


class HistoryStateMixin(object):
    """HTML5 history stae data mixin"""

    # option for skip browser history state
    skipState = False

    # history state title (not supported by all browser history implementations)
    stateTitle = ''

    # method for process state, by default we simply load the page content.
    # Note, by default the cbStateParams get used for calling this method
    # and if no cbStateParams get used they get extracted from the cbStateURL.
    # This means if there are any params, your method must accept the params
    # argument
    cbStateMethod = 'j01LoadContent'
    # optional params for cbStateMethod method, if None, the params get
    # extracted from the cbStateURL if any. Such cbStateParams must be json
    # seriazable
    cbStateParams = None
    # onSuccess callback used by state processing
    cbStateSuccess = 'j01RenderContentSuccess'
    # onError callback used by state processing
    cbStateError = 'j01RenderContentError'
    # onTimeout callback used by state processing
    cbStateTimeout = None

    @property
    def cbStateURL(self):
        """JSONRPC proxy url used for state processing

        NOTE: this url will force to load the content from the server via
        jsonrpc. By default the pageURL get used. If this cbStateURL is missing,
        the stateURL get used for load the page from the server using a non
        jsonrpc request. See stateURL description below for more information.
        """
        if not self.skipState:
            return self.pageURL

    @property
    def stateURL(self):
        """State URL (by default pageURL is used)

        This url is shown in browser navigation and also used for reload a page

        Note: this url will force to load the page with:

            window.location.href = url

        But only if the cbStateURL is not used. Because we first process a
        cbStateURL before we process the stateURL

        Note: Only use the stateURL url if the full page can get loaded within
        a browser request and make sure that the page provides a layout within
        the browser request. You can also use a totaly different url of  page
        which can get loaded including  layout if the page itself does not
        support a layout.

        Note: the stateURL is also shown in the browser navigation and will get
        use for navigation if the user clicks the browser relaod button e.g.
        ctrl/F5.

        Note; this url is also used if the user booksmarks a page.

        Remember, a jsonrpc request will only call update/render and a browser
        request will call __call__/update/render. Only the __call__ method will
        include the layout template.
        """
        if not self.skipState:
            return self.pageURL



##############################################################################
#
# Copyright (c) 2007 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
$Id:$
"""
from __future__ import absolute_import
__docformat__ = "reStructuredText"
from zope.interface import implementer

import unittest
import doctest

import persistent

import zope.interface
import zope.schema
import zope.location.interfaces
from zope.container import contained
from zope.schema.fieldproperty import FieldProperty

import z3c.form.field

#import z3c.jsonrpc.testing
from j01.jsonrpc import jsform
from j01.jsonrpc import testing


class IDemoContent(zope.location.interfaces.IContained):
    """Demo content interface."""

    title = zope.schema.TextLine(
        title=u'Title',
        description=u'The title')

    description = zope.schema.TextLine(
        title=u'Description',
        description=u'The description')


@implementer(IDemoContent)
class DemoContent(persistent.Persistent, contained.Contained):
    """Demo content."""

    title = FieldProperty(IDemoContent['title'])
    description = FieldProperty(IDemoContent['description'])


class DemoForm(jsform.JSONRPCEditForm):
    """Sample JSON form."""

    fields = z3c.form.field.Fields(IDemoContent).select(
        'title', 'description')


def test_suite():
    return unittest.TestSuite((
        testing.FunctionalDocFileSuite('README.txt',
            setUp=testing.setUp, tearDown=testing.tearDown,
            optionflags=doctest.NORMALIZE_WHITESPACE|doctest.ELLIPSIS,
            ),
        doctest.DocFileSuite('checker.txt',
            optionflags=doctest.NORMALIZE_WHITESPACE|doctest.ELLIPSIS),
        ))


if __name__=='__main__':
    unittest.main(defaultTest='test_suite')



#======
#README
#======

#JSON-RPC form processing
#------------------------

#This package offers forms which are able to process via JSON-RPC and other
#usefull JSON-RPC methods.

from __future__ import absolute_import
from __future__ import print_function
from pprint import pprint#>>>
import zope.interface#>>>
import zope.component#>>>
import zope.i18n.negotiator#>>>
import zope.traversing.testing#>>>
from zope.browserresource.resource import AbsoluteURL#>>>
from zope.interface import alsoProvides#>>>
from z3c.form.interfaces import IFormLayer#>>>
import z3c.form.testing#>>>
import j01.jsonrpc.tests#>>>
import j01.jsonrpc.testing#>>>
 from p01.publisher.testing import getRootFolder#>>>

#Setup negotiator used for translations:

zope.component.provideUtility(zope.i18n.negotiator.negotiator)#>>>

#Setup absolute URL adapter for resources, if available:

j01.jsonrpc.testing.setUpAbsoluteURL()#>>>
zope.component.provideAdapter(AbsoluteURL)#>>>

#Now we can setup a content container and content object in our site:

site  = getRootFolder()#>>>
content = j01.jsonrpc.tests.DemoContent()#>>>
site['content'] = content#>>>

#Now we can call the method from our JSONRPC view but first setup form support:

zope.traversing.testing.setUp()#>>>
z3c.form.testing.setupFormDefaults()#>>>
j01.jsonrpc.testing.setupJSONRPCFormDefaults()#>>>
formRequest = j01.jsonrpc.testing.FormTestRequest()#>>>

#Now test the JSON-RPC edit form:

demoEditForm = j01.jsonrpc.tests.DemoForm(content, formRequest)#>>>
demoEditForm.__name__ = u'demoEditForm'#>>>
demoEditForm.update()#>>>
print(demoEditForm.render())#>>>
#  <form action="http://127.0.0.1/content/demoEditForm" method="post" enctype="multipart/form-data" class="edit-form" id="form" name="form">
#    <div class="viewspace">
#        <div class="required-info">
#           <span class="required">*</span>&ndash; required
#        </div>
#      <div>
#            <div id="form-widgets-title-row" class="row required">
#                <div class="label">
#                  <label for="form-widgets-title">
#                    <span>Title</span>
#                    <span class="required">*</span>
#                  </label>
#                </div>
#                <div class="widget">
#      <input id="form-widgets-title" name="form.widgets.title" class="text-widget required textline-field" value="" type="text" />
#  <BLANKLINE>
#  </div>
#            </div>
#            <div id="form-widgets-description-row" class="row required">
#                <div class="label">
#                  <label for="form-widgets-description">
#                    <span>Description</span>
#                    <span class="required">*</span>
#                  </label>
#                </div>
#                <div class="widget">
#      <input id="form-widgets-description" name="form.widgets.description" class="text-widget required textline-field" value="" type="text" />
#  <BLANKLINE>
#  </div>
#            </div>
#      </div>
#    </div>
#    <div>
#      <div class="buttons">
#        <input type="button" id="form-buttons-applyChanges" name="form.buttons.applyChanges" class="button-widget" value="Apply" />
#  <BLANKLINE>
#  <script>
#              $('#form').on('click', '#form-buttons-applyChanges', function(){
#                  var $btn = $('#form-buttons-applyChanges');
#                  if (!$btn.prop('disabled')) {
#                      var data = $('#form').j01FormToArray('applyChanges');
#                      proxy = getJSONRPCProxy('http://127.0.0.1/content/demoEditForm');
#                      proxy.addMethod('j01FormProcessor', j01RenderContentSuccess, j01RenderContentError, null, null);
#                      proxy.j01FormProcessor(data);
#                  }
#                  return false;
#              });
#              </script>
#        <input type="button" id="form-buttons-cancel" name="form.buttons.cancel" class="button-widget" value="Cancel" />
#  <BLANKLINE>
#  <script>
#              $('#form').on('click', '#form-buttons-cancel', function(){
#                  var $btn = $('#form-buttons-cancel');
#                  if (!$btn.prop('disabled')) {
#                      var data = $('#form').j01FormToArray('cancel');
#                      proxy = getJSONRPCProxy('http://127.0.0.1/content/demoEditForm');
#                      proxy.addMethod('j01FormProcessor', j01RenderContentSuccess, j01RenderContentError, null, null);
#                      proxy.j01FormProcessor(data);
#                  }
#                  return false;
#              });
#              </script>
#      </div>
#    </div>
#  </form>
#  <BLANKLINE>


#As you can see the buttons get rendered including come javascript code. This
#code will call the form prcessing part. Let's try to send some JSON-RPC form
#varaible back to the server. but first we need to register the view as an
#adapter:

#Let's show how we can register a jsonrpc view for the container:
#(The container class needs permission configuration too)

from zope.configuration import xmlconfig#>>>
import p01.jsonrpc#>>>
import zope.security#>>>
import zope.component#>>>
import zope.publisher#>>>
import zope.browserpage#>>>
context = xmlconfig.file('meta.zcml', p01.jsonrpc)#>>>
context = xmlconfig.file('meta.zcml', zope.security, context)#>>>
context = xmlconfig.file('meta.zcml', zope.component, context)#>>>
context = xmlconfig.file('meta.zcml', zope.publisher, context)#>>>
context = xmlconfig.file('meta.zcml', zope.browserpage, context)#>>>
context = xmlconfig.string("""#>>>
<configure#...
    xmlns:z3c="http://namespaces.zope.org/z3c"#...
    xmlns="http://namespaces.zope.org/zope"#...
    xmlns:browser="http://namespaces.zope.org/browser">#...
    <class class="j01.jsonrpc.tests.DemoContent">#...
      <allow#...
          interface="j01.jsonrpc.tests.IDemoContent"#...
          />#...
    </class>#...
  <browser:page#...
      name="demoForm"#...
      for="j01.jsonrpc.tests.IDemoContent"#...
      class="j01.jsonrpc.tests.DemoForm"#...
      permission="zope.Public"#...
      layer="p01.jsonrpc.testing.IJSONRPCTestLayer"#...
      />#...
  <z3c:jsonrpc#...
      for="j01.jsonrpc.tests.DemoForm"#...
      class="j01.jsonrpc.jsonrpc.J01FormProcessor"#...
      permission="zope.Public"#...
      methods="j01FormProcessor"#...
      layer="p01.jsonrpc.testing.IJSONRPCTestLayer"#...
      />#...
</configure>#...
""", context)#...

import p01.jsonrpc.publisher#>>>
from zope.interface import classImplements#>>>
from p01.jsonrpc.testing import IJSONRPCTestSkin#>>>
classImplements(p01.jsonrpc.testing.IJSONRPCTestSkin, IFormLayer)#>>>

import os#>>>
import z3c.form.tests#>>>
import z3c.template.template#>>>
editFormTemplate = os.path.join(os.path.dirname(z3c.form.tests.__file__),#>>>
    'simple_edit.pt')#...
editTemplateFactory = z3c.template.template.TemplateFactory(#>>>
    editFormTemplate, 'text/html')#...

#We register the factory on a view interface and a layer.

import z3c.template.interfaces#>>>
import p01.jsonrpc.interfaces#>>>
zope.component.provideAdapter(editTemplateFactory,#>>>
    (zope.interface.Interface, p01.jsonrpc.interfaces.IJSONRPCRequest),#...
    z3c.template.interfaces.IContentTemplate)#...

from p01.jsonrpc.testing import JSONRPCTestProxy#>>>
proxy = JSONRPCTestProxy('http://globalmgr:globalmgrpw@127.0.0.1/++skin++JSONRPCTestSkin/content/demoForm')#>>>
data = {'j01FormHandlerName':'applyChanges',#>>>
        'form.widgets.title': u'Foo',#...
        'form.widgets.description': u'Bar',#...
        'form.buttons.applyChanges': u'Apply'}#...
result = proxy.j01FormProcessor(**data)#>>>
pprint(result)#>>>
#  {u'content': u'<!DOCTYPE html PUBLIC ...</html>\n',
#   u'contentRenderMethodName': None,
#   u'contentTargetExpression': None,
#   u'nextContentURL': None,
#   u'nextURL': None,
#   u'state': {u'cbURL': u'http://127.0.0.1/++skin++JSONRPCTestSkin/content/demoForm',
#              u'method': u'j01LoadContent',
#              u'onError': u'j01RenderContentError',
#              u'onSuccess': u'j01RenderContentSuccess',
#              u'onTimeout': None,
#              u'params': None,
#              u'title': u'',
#              u'url': u'http://127.0.0.1/++skin++JSONRPCTestSkin/content/demoForm'},
#   u'url': u'http://127.0.0.1/++skin++JSONRPCTestSkin/content/demoForm'}

print(result['contentTargetExpression'] is None)#>>>
#  True

print(result['content'])#>>>
#  <!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
#  <html xmlns="http://www.w3.org/1999/xhtml">
#    <body>
#      <form action=".">
#        <div class="row">
#          <label for="form-widgets-title">Title</label>
#      <span id="form-widgets-title" class="text-widget textline-field">Foo</span>
#  </div>
#        <div class="row">
#          <label for="form-widgets-description">Description</label>
#      <span id="form-widgets-description" class="text-widget textline-field">Bar</span>
#  </div>
#        <div class="action">
#  <input id="form-buttons-applyChanges" name="form.buttons.applyChanges" class="submit-widget jsonrpcbutton-field" value="Apply" type="submit" />
#  </div>
#        <div class="action">
#  <input id="form-buttons-cancel" name="form.buttons.cancel" class="submit-widget jsonrpcbutton-field" value="Cancel" type="submit" />
#  </div>
#      </form>
#    </body>
#  </html>



##############################################################################
#
# Copyright (c) 2002 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Resource base class and AbsoluteURL adapter
"""

from __future__ import absolute_import
from builtins import str
import zope.interface
import zope.component
import zope.component.hooks
from zope.traversing.browser.interfaces import IAbsoluteURL
import zope.traversing.browser.absoluteurl

import zope.browserresource.interfaces

from p01.jsonrpc.interfaces import IJSONRPCRequest


# XXX: move this to p01.jsonrpc, otherwise we can't lookup correct resource
#      path
class AbsoluteURL(zope.traversing.browser.absoluteurl.AbsoluteURL):

    zope.interface.implementsOnly(IAbsoluteURL)
    zope.component.adapts(zope.browserresource.interfaces.IResource,
        IJSONRPCRequest)

    def __init__(self, context, request):
        self.context = context
        self.request = request

    def _createUrl(self, baseUrl, name):
        return "%s/@@/%s" % (baseUrl, name)

    def __str__(self):
        name = self.context.__name__
        if name.startswith('++resource++'):
            name = name[12:]

        site = zope.component.hooks.getSite()
        base = zope.component.queryMultiAdapter((site, self.request),
            IAbsoluteURL, name="resource")
        if base is None:
            url = str(zope.component.getMultiAdapter((site, self.request),
                IAbsoluteURL))
        else:
            url = str(base)

        return self._createUrl(url, name)



##############################################################################
#
# Copyright (c) 2010 Projekt01 GmbH.
# All Rights Reserved.
#
##############################################################################
"""
$Id:$
"""

from __future__ import absolute_import
from builtins import object
import zope.interface
import zope.component
import zope.i18nmessageid
from zope.component import hooks
from zope.traversing.browser.absoluteurl import absoluteURL

from z3c.form.interfaces import IValue
from z3c.form.interfaces import IErrorViewSnippet

import p01.jsonrpc.error
from p01.jsonrpc.interfaces import IJSONRPCRequest

_ = zope.i18nmessageid.MessageFactory('p01')


class Unauthorized(p01.jsonrpc.error.JSONRPCErrorView):
    """Unauthorized error view which returns response.data.nextURL.

    The response can get handled by our j01.jsonrpc.js library and redirects to
    the given nextURL e.g. <site>/j01Login.html page.

    This is probably not the best concept because we will loose the
    original page and what the user was trying to do. But it's probably not
    possible to traverse to the jsonrpc method because of unauthorized.
    This seems to be the only solution whcih is working for any project. But
    feel free to implement a custom concept for you own project if you like
    to provide a better user experience.

    Note: the j01Login.html page is a simple browser page probably the same as
    your existing login page and not a jsonrpc method.

    Note: use your own Unauthorized error view if you don't like to redirect to
    the <site>/j01Login.html page.

    Since this unauthorized error handling concept doesn't know about the
    javascript library used on the client side you have to make sure that
    the given error response (response.data.nextURL) can get handled by the
    client side javascript. The j01.jsonrpc.js javascript used by this package
    supports this error handling. But make sure that your own custom javascript
    can handle such a response or replace this error view with oru own custom
    error response.
    """

    code = -32000
    message = u'Unauthorized'

    @property
    def data(self):
        # setup a redirect url
        site = hooks.getSite()
        nextURL = '%s/j01Login.html' % absoluteURL(site, self.request)
        return {'nextURL': nextURL}

    def __call__(self):
        """Must return itself.

        This allows us to use the error view in setResult if ZopePublication
        is adapting an error view to error and request and calls them.
        """
        # do not call unauthorized, it's useless since it will only setup a
        # redirect. JSON-RPC doens't provide such a redirect
        self.request.response.setStatus(401)
        return self


# z3c.form IErrorViewSnippet
@zope.component.adapter(zope.schema.ValidationError, IJSONRPCRequest, None,
    None, None, None)
@zope.interface.implementer(IErrorViewSnippet)
class JSONErrorViewSnippet(object):
    """Error view snippet used by z3c.form in jsonrpc request"""

    def __init__(self, error, request, widget, field, form, content):
        self.error = self.context = error
        self.request = request
        self.widget = widget
        self.field = field
        self.form = form
        self.content = content

    def update(self):
        value = zope.component.queryMultiAdapter(
            (self.context, self.request, self.widget,
             self.field, self.form, self),
            IValue, name='message')
        if value is not None:
            self.message = value.get()
        else:
            self.message = self.error.doc()

    def render(self):
        return self.message

    def __repr__(self):
        return '<%s for %s>' %(
            self.__class__.__name__, self.error.__class__.__name__)


@zope.component.adapter(ValueError, IJSONRPCRequest, None,
    None, None, None)
@zope.interface.implementer(IErrorViewSnippet)
class JSONValueErrorViewSnippet(JSONErrorViewSnippet):
    """An error view for ValueError raised by z3c.form widget validation."""

    @property
    def message(self):
        errMsg = _(u"The system could not process the given value.")
        return zope.i18n.translate(errMsg, context=self.request)

    def update(self):
        value = zope.component.queryMultiAdapter(
            (self.context, self.request, self.widget,
             self.field, self.form, self),
            IValue, name='message')
        if value is not None:
            self.message = value.get()

    def render(self):
        return self.message



##############################################################################
#
# Copyright (c) 2007 Projekt01 GmbH and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
$Id: __init__.py 6 2006-04-16 01:28:45Z roger.ineichen $
"""

from __future__ import absolute_import
import zope.schema
import zope.interface

import z3c.form.interfaces


###############################################################################
#
# action and handler concept

class IButtons(z3c.form.interfaces.IButtons):
    """Buttons."""

    def getInputEnterJavaScript(form, request):
        """Returns the input enter java script code if the inputEnterActionName
        name defines a button and the button condition is True.
        """


class IButtonHandlers(z3c.form.interfaces.IButtonHandlers):
    """A collection of handlers for buttons."""


class IButtonHandler(z3c.form.interfaces.IButtonHandler):
    """A handler managed by the button handlers."""

    def __call__(form, action):
        """Processes the action handler"""


class IActionHandler(z3c.form.interfaces.IActionHandler):

    def __call__():
        """Lookup and execute the action handler."""


###############################################################################
#
# action and widget

class IButtonWidget(z3c.form.interfaces.IButtonWidget):
    """Button widget."""


class IButtonAction(z3c.form.interfaces.IButtonAction):
    """Button Action."""

    javascript = zope.schema.Text(
        title=u'Javascript',
        description=(u'This attribute specifies the javascript part rendered '
                     u'directly after the button if given.'),
        default=None,
        required=False)

    data = zope.schema.Dict(
        title=u'Additional button data attributes',
        description=u'Additional button data attributes',
        default=None,
        required=False)


###############################################################################
#
# submit buttons

class IButton(z3c.form.interfaces.IButton):
    """Button using a string template and data attributes based on attrNames
    
    The default attribute names should be fine for most button implementations.
    Set the template to None in your own button if you like to use a page
    template for rendering the button. The simple string template allows us to
    render additional data attributes which you can define with a data dict
    or a dataGetter method. 
    """

    css = zope.schema.Field(
        title=u'CSS class',
        description=u'CSS class',
        required=False)

    template = zope.schema.TextLine(
        title=u'Button string template',
        description=u'Button string template',
        default=None,
        required=False)

    attrNames = zope.schema.Field(
        title=u'Attribute names used for button element',
        description=u'Attribute names used for button element',
        default=['type', 'id', 'name', 'value', 'class', 'title', 'alt',
                 'tabindex'],
        required=True)

    data = zope.schema.Dict(
        title=u'Additional button data attributes',
        description=u'Additional button data attributes',
        default=None,
        required=False)

    dataGetter = zope.interface.Attribute("""Button data getter method""")

    def getInputEnterJavaScript(form, request):
        """Returns the input enter JavaScript code."""

    def getJavaScript(action, request):
        """Returns the javascript code."""

    def getData(action, request, data):
        """Returns the the optional data-loading value.
        
        This value can get used as loading message after button click and is
        only supported in javascript and jsonrpc (ajax)  buttons.
        """


###############################################################################
#
# javascript buttons


class IJSButton(IButton):
    """JS javascript aware button."""

    urlGetter = zope.interface.Attribute("""URL getter method""")


class IJSONRPCButton(IJSButton):
    """JSONRPC button"""

    css = zope.interface.Attribute("""css""")

    callback = zope.schema.ASCIILine(
        title=u'Callback function name as string',
        description=u'Callback function name as string',
        default='',
        required=False)

    loading = zope.schema.TextLine(
        title=u'Button loading text',
        description=u'Button loading text',
        default=None,
        required=False)

    def getURL(form):
        """Returns the url based on urlGetter or the form url"""


###############################################################################
#
# jsonrpc page

class IJSONRPCPage(zope.interface.Interface):
    """JSON-RPC page marker."""


###############################################################################
#
# jsonrpc form

class IJSONRPCForm(zope.interface.Interface):
    """JSON-RPC base form mixin class."""

    inputEnterActionName = zope.schema.TextLine(
        title=(u'Button name where the input enter JS code get used'),
        description=(u'Button name where the input enter JS code get used'),
        required=False)

    inputEnterJavaScript = zope.schema.Text(
        title=u'Input enter javascript code',
        description=(u'This attribute specifies the javascript part rendered '
                     u'for bind input to keypress enter handler'),
        required=True)

    refreshWidgets = zope.schema.Bool(
        title=u'Refresh widgets',
        description=(u'A flag, when set, causes form widgets to be updated '
                     u'again after action execution.'),
        default=False,
        required=True)


class IJSONRPCAddForm(IJSONRPCForm, z3c.form.interfaces.IAddForm):
    """JSON-RPC based add form."""


class IJSONRPCEditForm(IJSONRPCForm, z3c.form.interfaces.IEditForm):
    """JSON-RPC based edit form."""


class IJSONRPCSearchForm(IJSONRPCForm, z3c.form.interfaces.IForm):
    """JSON-RPC based search form."""



##############################################################################
#
# Copyright (c) 2007 Projekt01 GmbH and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
$Id: __init__.py 6 2006-04-16 01:28:45Z roger.ineichen $
"""
from __future__ import absolute_import
from future import standard_library
from zope.interface import implementer
standard_library.install_aliases()
__docformat__ = "reStructuredText"

import sys
import urllib.request, urllib.parse, urllib.error

import zope.interface
import zope.component
import zope.i18nmessageid
from zope.traversing.browser import absoluteURL
from zope.publisher.interfaces import NotFound

import z3c.form.field
import z3c.form.form
from z3c.template.template import getPageTemplate
from z3c.template.template import getLayoutTemplate

from p01.jsonrpc.interfaces import IMethodPublisher

import j01.jsonrpc
from j01.jsonrpc import interfaces
from j01.jsonrpc import btn

_ = zope.i18nmessageid.MessageFactory('p01')


REDIRECT_STATUS_CODES = (301, 302, 303)


# supports z3c.form and j01.jsonrpc button and handlers
def extends(*args, **kwargs):
    """Copy form button, handler and fields from given form

    Note: this method supports both (z3c.form and j01.jsonrpc) concepts and
    uses the correct button and handlers.
    """
    frame = sys._getframe(1)
    f_locals = frame.f_locals
    if not kwargs.get('ignoreFields', False):
        f_locals['fields'] = z3c.form.field.Fields()
        for arg in args:
            f_locals['fields'] += getattr(arg, 'fields',
                z3c.form.field.Fields())
    if not kwargs.get('ignoreButtons', False):
        f_locals['buttons'] = btn.Buttons()
        for arg in args:
            f_locals['buttons'] += getattr(arg, 'buttons', btn.Buttons())
    if not kwargs.get('ignoreHandlers', False):
        f_locals['handlers'] = btn.Handlers()
        for arg in args:
            f_locals['handlers'] += getattr(arg, 'handlers', btn.Buttons())


@implementer(interfaces.IJSONRPCForm)
class JSONRPCFormMixin(j01.jsonrpc.LayoutHelperMixin,
    j01.jsonrpc.HistoryStateMixin):
    """JSONRPC form mixin providing IJSONRPCForm

    Note: the IJSONRPCForm interfaces makes sure that a pagelet
    (content provider) adapter is available by default. This is usefull if you
    instanciate such classes without to lookup them as adapters registred
    with the pagelet directive.
    """


    template = getPageTemplate()
    layout = getLayoutTemplate()

    formErrorsMessage = _('There were some errors.')
    successMessage = _('Data successfully updated.')
    noChangesMessage = _('No changes were applied.')

    # cached urls
    _contextURL = None
    _pageURL = None

    # override widget prefix if you need to load different forms using the same
    # field names in one single page.
    prefixWidgets = 'widgets'

    # override button prefix if you need to load different forms using the same
    # button in one single page.
    prefixButtons = 'buttons'

    # allows to skip action and widget processing. This is sometimes required
    # for JSON-RPC forms
    skipActions = False
    skipWidgets = False

    # allows to fetch a status message from an url set by redirect. Set this to
    # None or a different name if you don't like to support such status
    # messages e.g. foo.html?status=foobar. Note: such status message must
    # be urlencoded
    statusAttrName = 'status'

    # set this conditions in your action handler method if needed
    # widgets normaly not change their value
    refreshWidgets = False
    # action condition may have changed after action execution
    refreshActions = False

    inputEnterActionName = None # see inputEnterJavaScript

    # JSON-RPC javascript callback arguments
    # content target expression where the result get rendered in. By default
    # ``#content`` get used if None is given
    contentTargetExpression = None

    # JSON-RPC content render javascript method name
    # The method registered as j01ContentRenderMethod can get used if the
    # default contentTargetExpression concept is not enough. Register a
    # javascript method with the j01.jsonrpc.js method j01RegisterRenderCallback
    # and we will lookup such a method with the method j01GetRenderCallback
    # in the default onSuccess handler. Or just use your own onSuccess callback
    # handler and use the contentRenderMethodName for invoke whatever you need
    contentRenderMethodName = None

    # optional scrollTo expression where the callback method will scroll to
    # after rendering. The default implementation uses offset().top. Feel free
    # to implement a custom j01CallbackRegistry method which uses another
    # concept instead of add more callback arguments
    scrollToExpression = None
    scrollToOffset = None
    scrollToSpeed = None

    # the next URL where the jsonrpc callback method will redirect to using
    # window.location.href = response.nextURL
    nextURL = None
    # the nextHash will update the url without to redirect
    # Note: we don't provide any javascript for hasbang since we fully support
    # the browser history api. Implement your own navigation concept if you
    # need it
    nextHash = None

    # url used for load content via j01LoadContent method
    nextContentURL = None

    buttons = btn.Buttons()
    handlers = btn.Handlers()

    def publishTraverse(self, request, name):
        view = zope.component.queryMultiAdapter((self, request), name=name)
        if view is None or not IMethodPublisher.providedBy(view):
            raise NotFound(self, name, request)
        return view

    @property
    def action(self):
        """Take care on action url."""
        return self.pageURL

    @property
    def contextURL(self):
        """Setup and cache context URL"""
        if self._contextURL is None:
            self._contextURL = absoluteURL(self.context, self.request)
        return self._contextURL

    @property
    def pageURL(self):
        """Setup and cache context URL"""
        if self._pageURL is None:
            self._pageURL = '%s/%s' % (absoluteURL(self.context, self.request),
                self.__name__)
        return self._pageURL

    @property
    def inputEnterJavaScript(self):
        """Enter button click handler.

        You can define an action handler name which get called on enter button
        call in your form like:

        inputEnterActionName = 'myHandlerName'

        Note: you need to include the inputEnter javascript in your template
        within:

        <script type="text/javascript"
                tal:content="view/inputEnterJavaScript">
        </script>
        """
        return self.buttons.getInputEnterJavaScript(self, self.request)

    def updateActions(self):
        if self.prefixButtons is not None:
            # overrride button prefix before update actions
            self.buttons.prefix = self.prefixButtons
        super(JSONRPCFormMixin, self).updateActions()

    def updateWidgets(self, prefix=None):
        if prefix is None and self.prefixWidgets is not None:
            prefix = self.prefixWidgets
        super(JSONRPCFormMixin, self).updateWidgets(prefix)

    def executeActions(self):
        """Dispatch actions.execute call"""
        self.actions.execute()

    def update(self):
        """Update form

        The default z3c.form calles the following methods in update:

        self.updateWidgets()
        self.updateActions()
        self.actions.execute()
        if self.refreshActions:
            self.updateActions()

        We implemented the following coditions:

        - skipWidgets
        - skipActions
        - refreshActions (also supported by z3c.form)
        - refreshWidgets

        This allows us to prepare the JSONRPC call setup and gives more
        controll for complex form setup. Also see J01FormProcessor in
        j01/jsform/jsonrpc.py

        """
        if not self.skipWidgets:
            # default False
            self.updateWidgets()

        if not self.skipActions:
            # default False
            self.updateActions()
            self.executeActions()

        if self.refreshActions:
            # default False
            self.updateActions()

        if self.refreshWidgets:
            # default False
            self.updateWidgets()

        if self.statusAttrName is not None:
            # get and set status given from url
            status = self.request.get(self.statusAttrName)
            if status:
                self.status = status

    def setNextURL(self, url, status):
        """Helper for set a nextURL including a status message

        Note: the status message must be an i18n message which will get
        translated later as status message.

        If you don't use a status message just use self.nextURL = myURL and
        don't use this method.

        """
        self.nextURL = '%s?%s' % (url, urllib.parse.urlencode({'status':status}))

    def render(self):
        if self.nextURL is not None:
            return None
        return self.template()

    def __call__(self):
        # don't render on redirect
        if self.request.response.getStatus() in REDIRECT_STATUS_CODES:
            return u''
        self.update()
        if self.request.response.getStatus() in REDIRECT_STATUS_CODES:
            return u''
        # we only use the nextURL pattern and not the redirect status check
        # in our JSON-RPC method call. But this __call__ method get only used
        # if no JSON-RPC call is involved. So let's setup the redirect.
        if self.nextURL is not None:
            self.request.response.redirect(self.nextURL)
            return u''
        return self.layout()


class JSONRPCForm(JSONRPCFormMixin, z3c.form.form.Form):
    """JSONRPC form mixin."""

    buttons = btn.Buttons()


@implementer(interfaces.IJSONRPCEditForm)
class JSONRPCEditForm(JSONRPCFormMixin, z3c.form.form.EditForm):
    """JSONRPC edit form."""


    showCancel = True
    buttons = btn.Buttons()

    def doHandleApplyChanges(self, action):
        # Note we, use the data from the request.form
        data, errors = self.extractData()
        if errors:
            self.status = self.formErrorsMessage
            return
        changes = self.applyChanges(data)
        if changes:
            self.status = self.successMessage
        else:
            self.status = self.noChangesMessage
        return changes

    def doHandleCancel(self, action):
        self.ignoreRequest = True
        self.updateActions()
        self.updateWidgets()

    @btn.buttonAndHandler(btn.IJSONRPCButtons['applyChanges'])
    def handleApplyChanges(self, action):
        self.doHandleApplyChanges(action)

    @btn.buttonAndHandler(btn.IJSONRPCButtons['cancel'],
        condition=lambda form: form.showCancel)
    def handleCancel(self, action):
        self.doHandleCancel(action)


@implementer(interfaces.IJSONRPCAddForm)
class JSONRPCAddForm(JSONRPCFormMixin, z3c.form.form.AddForm):
    """JSONRPC add form."""


    showCancel = True
    buttons = btn.Buttons()

    def doHandleAdd(self, action):
        data, errors = self.extractData()
        if errors:
            self.status = self.formErrorsMessage
            return
        obj = self.createAndAdd(data)
        if obj is not None:
            # mark only as finished if we get the new object
            self._finishedAdd = True

    def doHandleCancel(self, data):
        self.ignoreRequest = True
        self.updateActions()
        self.updateWidgets()

    @btn.buttonAndHandler(btn.IJSONRPCButtons['add'])
    def handleAdd(self, action):
        self.doHandleAdd(action)

    @btn.buttonAndHandler(btn.IJSONRPCButtons['cancel'],
        condition=lambda form: form.showCancel)
    def handleCancel(self, action):
        self.doHandleCancel(action)

    def renderAfterAdd(self):
        return super(JSONRPCAddForm, self).render()

    def render(self):
        if self._finishedAdd:
            return self.renderAfterAdd()
        return super(JSONRPCAddForm, self).render()


@implementer(interfaces.IJSONRPCSearchForm)
class JSONRPCSearchForm(JSONRPCFormMixin, z3c.form.form.Form):
    """JSONRPC search form."""


    buttons = btn.Buttons()
    inputEnterActionName = 'search'

    def doHandleSearch(self, action):
        raise NotImplementedError('Subclass must implement doHandleSearch')

    @btn.buttonAndHandler(btn.IJSONRPCButtons['search'])
    def handleSearch(self, action):
        self.doHandleSearch(action)



##############################################################################
#
# Copyright (c) 2007 Projekt01 GmbH and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
$Id: __init__.py 6 2006-04-16 01:28:45Z roger.ineichen $
"""

from __future__ import absolute_import
import zope.interface
from zope.interface import implementer
from zope.traversing.browser import absoluteURL

import z3c.pagelet.browser
from z3c.template.template import getPageTemplate
from z3c.template.template import getLayoutTemplate

from p01.jsonrpc.browser import JSONRPCTraversablePage

import j01.jsonrpc
from j01.jsonrpc import interfaces

REDIRECT_STATUS_CODES = (301, 302, 303)


@implementer(interfaces.IJSONRPCPage)
class JSONRPCPage(JSONRPCTraversablePage, j01.jsonrpc.LayoutHelperMixin,
    j01.jsonrpc.HistoryStateMixin, z3c.pagelet.browser.BrowserPagelet):
    """JSON-RPC traversable pagelet.

    Note: this JSON-RPC pagelet is also usable as normal BrowserPagelet
    replacement. The pagelet only supports an additional traversal option.
    see: JSONRPCTraversablePage
    This traversal option should not be a problem since we need the permission
    to traverse the page first before we traverse to the JSON-RPC handler.
    """


    layout = getLayoutTemplate()
    template = getPageTemplate()

    _contextURL = None
    _pageURL = None

    # JSON-RPC javascript callback arguments
    # content target expression where the result get rendered in. By default
    # the built-in argument ``#content`` get used
    contentTargetExpression = None

    # JSON-RPC content render javascript method name
    # The method registered as j01ContentRenderMethod can get used if the
    # default contentTargetExpression concept is not enough. Register a
    # javascript method with the j01.jsonrpc.js method j01RegisterRenderCallback
    # and we will lookup such a method with the method j01GetRenderCallback
    # in the default onSuccess handler. Or just use your own onSuccess callback
    # handler and use the contentRenderMethodName for invoke whatever you need
    contentRenderMethodName = None

    # optional scrollTo expression where the callback method will scroll to
    # after rendering. The default implementation uses offset().top. Feel free
    # to implement a custom j01CallbackRegistry method which uses another
    # concept instead of add more callback arguments
    scrollToExpression = None
    scrollToOffset = None
    scrollToSpeed = None

    # the next URL where the jsonrpc callback method will redirect to using
    # window.location.href = response.nextURL
    nextURL = None
    # the nextHash will update the url witout to redirect
    nextHash = None

    # url used for load content via j01LoadContent method
    nextContentURL = None

    @property
    def contextURL(self):
        """Setup and cache context URL"""
        if self._contextURL is None:
            self._contextURL = absoluteURL(self.context, self.request)
        return self._contextURL

    @property
    def pageURL(self):
        """Setup and cache context URL"""
        if self._pageURL is None:
            self._pageURL = '%s/%s' % (absoluteURL(self.context, self.request),
                self.__name__)
        return self._pageURL

    def render(self):
        if self.nextURL is not None:
            return None
        return self.template()

    def __call__(self):
        self.update()
        if self.nextURL is not None:
            self.request.response.redirect(self.nextURL)
            return u''
        return self.layout()



##############################################################################
#
# Copyright (c) 2007 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Setup

$Id:$
"""
from __future__ import absolute_import
import os
from setuptools import setup, find_packages

def read(*rnames):
    return open(os.path.join(os.path.dirname(__file__), *rnames)).read()

setup (
    name='j01.jsonrpc',
    version='3.0.0',
    author = "Roger Ineichen, Projekt01 GmbH",
    author_email = "dev@projekt01.ch",
    description = "JSON-RPC helpers based on JQuery, z3c.form and z3cjsonrpc for Zope 3",
    long_description=(
        read('README.txt')
        + '\n\n' +
        read('CHANGES.txt')
        ),
    license = "ZPL 2.1",
    keywords = "J01 XHR ajax z3c form JSON-RPC JQuery Zope3",
    classifiers = [
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Zope Public License',
        'Programming Language :: Python',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Topic :: Internet :: WWW/HTTP',
        'Framework :: Zope3'],
    url = 'http://pypi.python.org/pypi/j01.jsonrpc',
    packages = find_packages('src'),
    include_package_data = True,
    package_dir = {'':'src'},
    namespace_packages = ['j01'],
    extras_require = dict(
        test = [
            'beautifulsoup4',
            'p01.checker',
            'p01.testbrowser',
            'p01.jsonrpc[test]',
            'z3c.form[test]',
            'z3c.macro',
            'z3c.template',
            'zope.container',
            'zope.pagetemplate',
            'zope.password',
            'zope.publisher',
            'zope.schema',
            'zope.testing',
            'zope.traversing',
            'zope.app.testing',
            ],
        ),
    install_requires = [
        'setuptools',
        'p01.jsonrpc',
        'z3c.form',
        'z3c.pagelet',
        'z3c.template',
        'zope.browserresource',
        'zope.component',
        'zope.i18n',
        'zope.i18nmessageid',
        'zope.interface',
        'zope.location',
        'zope.publisher',
        'zope.schema',
        'zope.site',
        'zope.traversing',
        ],
    zip_safe = False,
    )



###############################################################################
#
# Copyright (c) 2014 Projekt01 GmbH
# All Rights Reserved.
#
###############################################################################
"""Buttons
$Id: btn.py 5189 2025-03-27 02:29:10Z felipe.souza $
"""
from __future__ import absolute_import
from six import string_types
from builtins import object
import six
__docformat__ = "reStructuredText"

import os
import sys

import zope.schema
import zope.interface
import zope.component
import zope.i18n
import zope.i18nmessageid
import zope.location
import zope.interface.adapter
from zope.pagetemplate.interfaces import IPageTemplate
from zope.schema.fieldproperty import FieldProperty

import z3c.form.interfaces
import z3c.form.button
import z3c.form.action
import z3c.form.widget
import z3c.form.util
import z3c.form.browser.widget

from j01.jsonrpc import interfaces

_ = zope.i18nmessageid.MessageFactory('p01')


try:
    str
except NameError:
    # Py3: Define unicode.
    str = str


###############################################################################
#
# decorators

def handler(btnOrName):
    """A decorator for defining a success handler."""
    if isinstance(btnOrName, six.string_types):
        __name__ = btnOrName
    else:
        __name__ = btnOrName.__name__
    def createHandler(func):
        handler = Handler(func)
        frame = sys._getframe(1)
        f_locals = frame.f_locals
        handlers = f_locals.setdefault('handlers', Handlers())
        handlers.addHandler(__name__, handler)
        return handler
    return createHandler


def buttonAndHandler(btnOrTitle, *args, **kwargs):
    """Button and handler setup decorator"""
    if isinstance(btnOrTitle, six.string_types):
        # add the title to button constructor keyword arguments
        kwargs['title'] = btnOrTitle
        # create button and add it to the button manager
        button = Button(**kwargs)
    else:
        button = btnOrTitle
    if args:
        __name__ = args[0]
    else:
        __name__ = kwargs.get('name', button.__name__)
    for k, v in list(kwargs.items()):
        if k == 'name':
            continue
        # apply additional attributes
        setattr(button, k, v)
    # Extract directly provided interfaces:
    provides = kwargs.pop('provides', ())
    if provides:
        zope.interface.alsoProvides(button, provides)
    frame = sys._getframe(1)
    f_locals = frame.f_locals
    buttons = f_locals.setdefault('buttons', Buttons())
    f_locals['buttons'] += Buttons(button)
    # return create handler method
    return handler(button and button or __name__)


###############################################################################
#
# action and handler concept

import zope.interface.interfaces

# IButtons
@zope.interface.implementer(interfaces.IButtons)
class Buttons(z3c.form.button.Buttons):
    """Button manager."""

    def __init__(self, *args):
        buttons = []
        for arg in args:
            if zope.interface.interfaces.IInterface.providedBy(arg):
                for name, button in zope.schema.getFieldsInOrder(arg):
                    if interfaces.IButton.providedBy(button):
                        buttons.append((name, button))
            elif self.managerInterface.providedBy(arg):
                buttons += list(arg.items())
            elif interfaces.IButton.providedBy(arg):
                if not arg.__name__:
                    arg.__name__ = z3c.form.util.createId(arg.title)
                buttons.append((arg.__name__, arg))
            else:
                raise TypeError("Unrecognized argument type", arg)
        super(z3c.form.util.SelectionManager, self).__init__(buttons)
        keys = []
        seq = []
        byname = {}
        for name, button in buttons:
            keys.append(name)
            seq.append(button)
            byname[name] = button

        self._data_keys = keys
        self._data_values = seq
        self._data = byname

    def getInputEnterJavaScript(self, form, request):
        """Returns the input enter java script code if the inputEnterActionName
        name defines a button and the button condition is True.
        """
        # find and return the form submit javascript
        btnName = getattr(form, 'inputEnterActionName', None)
        if btnName is not None:
            button = self.get(btnName)
            # note button AND condition could be None
            if button is not None:
                if button.condition is None or (
                    button.condition is not None and button.condition(form)):
                    return button.getInputEnterJavaScript(form, request)

    def __repr__(self):
        return "<%s %r>" % (self.__class__.__name__, list(self._data.keys()))


@zope.interface.implementer(interfaces.IButtonHandlers)
class Handlers(object):
    """Action Handlers for a Button-based form."""

    def __init__(self):
        # setup name, handler container
        self._data = {}

    def addHandler(self, name, handler):
        self._data[name] = handler

    def getHandler(self, name, default=None):
        return self._data.get(name, default)

    def copy(self):
        handlers = Handlers()
        for name, handler in list(self._data.items()):
            handlers.addHandler(name, handler)
        return handlers

    def __add__(self, other):
        """See interfaces.IButtonHandlers"""
        if not isinstance(other, Handlers):
            raise NotImplementedError
        handlers = self.copy()
        for name, handler in list(other._data.items()):
            handlers.addHandler(name, handler)
        return handlers

    def __repr__(self):
        return "<%s %r>" % (self.__class__.__name__, list(self._data.keys()))


@zope.interface.implementer(interfaces.IButtonHandler)
class Handler(object):
    """Handler handler."""

    def __init__(self, func):
        self.func = func

    def __call__(self, form, action):
        return self.func(form, action)

    def __repr__(self):
        return '<%s for %r>' %(self.__class__.__name__, self.func.__name__)


@zope.interface.implementer(interfaces.IActionHandler)
class ActionHandler(object):
    """Button handler executer.

    This adapter makes it possible to execute button handler.
    """

    zope.component.adapts(
        z3c.form.interfaces.IForm,
        zope.interface.Interface,
        zope.interface.Interface,
        z3c.form.interfaces.IButtonAction)

    def __init__(self, form, request, content, action):
        self.form = form
        self.request = request
        self.content = content
        self.action = action

    def __call__(self):
        handler = self.form.handlers.getHandler(self.action.__name__)
        if handler is None:
            return
        return handler(self.form, self.action)


###############################################################################
#
# button action widget

@zope.interface.implementer(interfaces.IButtonAction)
@zope.component.adapter(z3c.form.interfaces.IFormLayer, interfaces.IButton)
class ButtonAction(z3c.form.action.Action,
    z3c.form.browser.widget.HTMLInputWidget, z3c.form.widget.Widget,
    zope.location.Location):
    """A button action supporting javascript"""

    klass = u'button-widget'

    def __init__(self, request, field):
        z3c.form.action.Action.__init__(self, request, field.title)
        z3c.form.widget.Widget.__init__(self, request)
        self.field = field

    def isExecuted(self):
        j01FormHandlerName = self.request.get('j01FormHandlerName')
        if j01FormHandlerName and self.name.endswith(j01FormHandlerName):
            return True
        else:
            # also support non JSONRPC request concept for urls like
            # <page-url>?form.buttons.foobar=1
            return self.name in self.request

    @property
    def _type(self):
        return self.field._type

    @property
    def accesskey(self):
        return self.field.accessKey

    @property
    def value(self):
        # translated title used as button value
        return zope.i18n.translate(self.title, context=self.request)

    @property
    def id(self):
        return self.name.replace('.', '-')

    # access css from button
    @property
    def css(self):
        return self.field.css

    @property
    def javascript(self):
        return self.field.getJavaScript(self, self.request)

    @property
    def data(self):
        """Returns relevant widget data used for button template"""
        # setup template data
        data = {
            'type': self._type,
            'id': self.id,
            'name': self.name,
            'value': self.value,
            'class': self.css and self.css or self.klass,
            }
        # apply additional template data defined in button
        for attrName in self.field.attrNames:
            # get attribute values from button
            v = getattr(self.field, attrName, None)
            if v:
                data[attrName] = v
        # apply button data
        btnData = self.field.getData(self, self.request)
        if btnData:
            dstr = ''
            for k, v in list(btnData.items()):
                dstr += 'data-%s="%s" ' % (k, v)
            data['data'] = dstr
        else:
            data['data'] = ''
        return data

    def update(self):
        """skip widget update call, there is nothing to do for a button"""
        pass

    def render(self):
        """Render the plain widget without additional layout"""
        if self.field.template is not None:
            # use the button string template providing data attributes
            javascript = self.javascript
            if javascript:
                javascript = '\n<script>%s</script>' % javascript
            else:
                javascript = ''
            return '%s%s' % (self.field.template % self.data, javascript)
        else:
            # use the page template (only used if template=None get used in
            # button setup
            template = self.template
            if template is None:
                template = zope.component.getMultiAdapter(
                    (self.context, self.request, self.form, self.field, self),
                    IPageTemplate, name=self.mode)
            return template(self)


###############################################################################
#
# submit buttons

BUTTON = """<input type="%(type)s" id="%(id)s" name="%(name)s" class="%(class)s" value="%(value)s" %(data)s/>
"""

@zope.interface.implementer(interfaces.IButton)
class Button(zope.schema.Field):
    """A button with a custom css class attribute"""

    accessKey = FieldProperty(interfaces.IButton['accessKey'])
    actionFactory = FieldProperty(interfaces.IButton['actionFactory'])
    css = FieldProperty(interfaces.IButton['css'])

    _type = 'submit'

    # used attribute names
    attrNames = [
        'tabindex'
        'title',
        'alt',
        ]

    data = None
    dataGetter = None
    template = BUTTON

    def __init__(self, *args, **kwargs):
        # Provide some shortcut ways to specify the name
        if args:
            kwargs['__name__'] = args[0]
            args = args[1:]
        if 'name' in kwargs:
            kwargs['__name__'] = kwargs['name']
            del kwargs['name']
        # apply additonal data key/values
        if 'template' in kwargs:
            self.template = kwargs['template']
            del kwargs['template']
        # apply additonal data key/values
        if 'data' in kwargs:
            self.data = kwargs['data']
            del kwargs['data']
        # apply optional dataGetter
        if 'dataGetter' in kwargs:
            self.dataGetter = kwargs['dataGetter']
            del kwargs['dataGetter']
        # apply optional css, which get added in front of other classes
        if 'css' in kwargs:
            self.css = kwargs['css']
            del kwargs['css']
        # Extract button-specific arguments
        self.accessKey = kwargs.pop('accessKey', None)
        self.condition = kwargs.pop('condition', None)
        # Initialize the button
        super(Button, self).__init__(*args, **kwargs)

    def getInputEnterJavaScript(self, form, request):
        """A simple button doesn't provide any javascript"""
        return u""

    def getJavaScript(self, action, request):
        """A simple button doesn't provide any javascript"""
        return u""

    def getData(self, action, request):
        """Returns additional data attibutes"""
        data = {}
        if self.data is not None:
            data.update(self.data)
        if self.dataGetter is not None:
            data.update(self.dataGetter(action, request))
        return data

    def __repr__(self):
        return '<%s %r %r>' %(self.__class__.__name__, self.__name__,
            self.title)


###############################################################################
#
# javascript buttons

@zope.interface.implementer(interfaces.IJSButton)
class JSButton(Button):
    """JS button.

    This is the basic implementation and only shows an alert message. Use this
    class for implement your own custom buttons.
    """

    _type = 'button'
    urlGetter = None

    def __init__(self, *args, **kwargs):
        # apply optional urlGetter
        if 'urlGetter' in kwargs:
            self.urlGetter = kwargs['urlGetter']
            del kwargs['urlGetter']
        super(JSButton, self).__init__(*args, **kwargs)

    def getURL(self, form, request):
        """Returns the url based on urlGetter or the form url"""
        if self.urlGetter is not None:
            return self.urlGetter(form)
        else:
            return form.pageURL

    def getInputEnterJavaScript(self, form, request):
        raise NotImplementedError(
            "Subclass must implement getInputEnterJavaScript method")

    def getJavaScript(self, action, request):
        raise NotImplementedError(
            "Subclass must implement getJavaScript method")


# JQuery trigger event
TRIGGER_JAVASCRIPT = """
                $btn.trigger({
                    type: '%(type)s',
                    form: '%(form)s',
                    action: '%(action)s',
                    handler: '%(handler)s'
                });"""

class TriggerMixin(object):
    """JQuery event trigger mixin

    Include j01.form.button.js for support button loading messages and button
    disable state.

    Enable JQuery event trigger with default j01.form.button.click event
    or implement your own event type using the trigger attribute. You can also
    add more event attributes by implement your own trigger call using the
    jsTrigger attribute.
    """

    loading = None
    trigger = None
    jsTrigger = None
    defaultTriggerEventType = 'j01.form.button.click'

    def __init__(self, *args, **kwargs):
        # apply button loading setup
        if 'loading' in kwargs:
            # setup i18n loading message
            self.loading = kwargs['loading']
            del kwargs['loading']
        # apply button trigger setup (also required for loading conecpt)
        if 'trigger' in kwargs:
            self.trigger = kwargs['trigger']
            del kwargs['trigger']
            if self.trigger is True:
                # trigger arg was True, use default trigger event type
                self.trigger = self.defaultTriggerEventType
            self.jsTrigger = TRIGGER_JAVASCRIPT
        if 'jsTrigger' in kwargs:
            # allow to override the javascript
            self.jsTrigger = kwargs['jsTrigger']
            del kwargs['jsTrigger']
        super(TriggerMixin, self).__init__(*args, **kwargs)

    def getData(self, action, request):
        data = super(TriggerMixin, self).getData(action, request)
        if self.loading:
            data['j01-loading-text'] = zope.i18n.translate(self.loading,
                context=request)
        return data

    def getTrigger(self, action):
        if self.jsTrigger:
            return self.jsTrigger % {
                'type': self.trigger,
                'form': action.form.id.replace('.', '\\\.'),
                'action': action.id,
                'handler': self.__name__,
                }
        else:
            return ''


# form processing button
@zope.interface.implementer(interfaces.IJSONRPCButton)
class JSONRPCButton(TriggerMixin, JSButton):
    """JSON-RPC for msubmit button

    This button will submit a form using the JSON-RPC method j01FormProcessor
    and renders the content into the existing page dom using the javascript
    method j01RenderContent.

    Only use this button if the form is a part of the loading JSON-RPC content.
    If the form tag is not a part of the loading JSON-RPC content them button
    javascript will apply it's handler to the form again and again which will
    end in multiple form submits on click. Use the JSONRPCClickButton if the
    form tag is not a part of the loading JSON-RPC content.

    Redirect is also supported if the form will set a nextURL. Or you can
    define a location where the content should get rendered by define a
    contentTargetExpression in your page/form.

    The getInputEnterJavaScript method get called if you set the buttons
    action handler name as inputEnterActionName value e.g:

    inputEnterActionName = 'applyChanges'

    Don't forget to include the inputEnterJavaScript part in your form
    template if you use an inputEnterActionName e.g.:

    <script type="text/javascript"
            tal:content="view/inputEnterJavaScript"> </script>

    This button requires the following javascript files:

    - jquery.js >= 1.7.0
    - j01.proxy.js
    - j01.jsonrpc.js

    NOTE: The form using this button must contain the form tag. If the form
    tag doesn't get rendered within each new request, the button click handler
    will get applied more then once to the form and will submit the form more
    then once.

    Note: file uploads can't get handeled by JSON-RPC or ajax. You have to
    use an iframe. See j01.dialog for built in iframe support.

    Note: the rendered button is supported by p01.testbrowser testing!
    You can test this with something like:

        button = browser.getControl(label=None, name=None, index=None)
        button.click()

    """

    typ = 'JSONRPCButton'
    method = 'j01FormProcessor'
    onSuccess = 'j01RenderContentSuccess'
    onError = 'j01RenderContentError'
    onTimeout = None
    isPushState = None
    idGetter = None

    def __init__(self, *args, **kwargs):
        # conditions
        if 'onSuccess' in kwargs and 'callback' in kwargs:
            raise ValueError(
                "Can't use onSuccess and callback argument for button setup")
        if 'onError' in kwargs and 'callback' in kwargs:
            raise ValueError(
                "Can't use onError and callback argument for button setup")
        # apply callback methods
        if 'onSuccess' in kwargs:
            self.onSuccess = kwargs['onSuccess']
            del kwargs['onSuccess']
        if 'onError' in kwargs:
            self.onError = kwargs['onError']
            del kwargs['onError']
        if 'onTimeout' in kwargs:
            self.onTimeout = kwargs['onTimeout']
            del kwargs['onTimeout']
        if 'callback' in kwargs:
            self.onSuccess = kwargs['callback']
            self.onError = kwargs['callback']
            del kwargs['callback']
        # apply isPushState
        if 'isPushState' in kwargs:
            self.isPushState = kwargs['isPushState']
            del kwargs['isPushState']
        # apply jsonrpc request id getter method
        if 'idGetter' in kwargs:
            self.idGetter = kwargs['idGetter']
            del kwargs['idGetter']

        # apply loading message
        if 'loading' in kwargs:
            self.loading = kwargs['loading']
            del kwargs['loading']
        super(JSONRPCButton, self).__init__(*args, **kwargs)

    def getData(self, action, request):
        """Returns additional data attibutes and the testing control type

        NOTE: this button data attributes are used for testing. The Browser
        located in p01.testbrowser.browser is able to handle the button clicks
        using python testing hook methods.
        """
        data = super(JSONRPCButton, self).getData(action, request)
        if request.get('paste.testing'):
            data['j01-testing-typ'] = self.typ
            data['j01-testing-url'] = self.getURL(action.form, request)
            data['j01-testing-form'] = action.form.name
            data['j01-testing-method'] = self.method
            data['j01-testing-error'] = self.onError
            data['j01-testing-success'] = self.onSuccess
            if self.onTimeout:
                data['j01-testing-timeout'] = self.onTimeout
        return data

    def getMethodArguments(self, request):
        """Returns the method arguments (don't skip one, except the last)"""
        # method, onSuccess, onError
        res = "'%s', %s, %s" % (self.method, self.onSuccess, self.onError)
        # onTimeout
        if self.onTimeout:
            res  += ", %s" % self.onTimeout
        else:
            res += ", null"
        # isPushState
        if self.isPushState is True:
            res  += ", true"
        elif self.isPushState is False:
            res  += ", false"
        else:
            res += ", null"
        # id
        if self.idGetter is not None:
            id = self.idGetter(request)
            if id is not None:
                res  += ", '%s'" % id
        return res

    def getInputEnterJavaScript(self, form, request):
        """Returns the input enter JavaScript code"""
        # replace dots in id with '\\.' See jquery.com for details
        return """
            $('#%(form)s').on('keypress', 'input', function(e){
                if(!e){e = window.event;}
                key = e.which ? e.which : e.keyCode;
                if (key == 13) {
                    var data = $('#%(form)s').j01FormToArray('%(handler)s');
                    proxy = getJSONRPCProxy('%(url)s');
                    proxy.addMethod(%(arguments)s);
                    proxy.%(method)s(data);
                    return false;
                }
            });
            """ % {'form': form.id.replace('.', '\\\.'),
                   'url': self.getURL(form, request),
                   'handler': self.__name__,
                   'method': self.method,
                   'arguments': self.getMethodArguments(request),
                   }

    def getJavaScript(self, action, request):
        # replace dots in id with '\\.' See jquery.com for details
        return """
            $('#%(form)s').on('click', '#%(action)s', function(){
                var $btn = $('#%(action)s');
                if (!$btn.prop('disabled')) {%(trigger)s
                    var data = $('#%(form)s').j01FormToArray('%(handler)s');
                    proxy = getJSONRPCProxy('%(url)s');
                    proxy.addMethod(%(arguments)s);
                    proxy.%(method)s(data);
                }
                return false;
            });
            """ % {'form': action.form.id.replace('.', '\\\.'),
                   'url': self.getURL(action.form, request),
                   'action': action.id,
                   'handler': self.__name__,
                   'trigger': self.getTrigger(action),
                   'method': self.method,
                   'arguments': self.getMethodArguments(request),
                   }


@zope.interface.implementer(interfaces.IJSONRPCButton)
class JSONRPCClickButton(JSONRPCButton):
    """JSON-RPC for msubmit button

    This button will submit a form using the JSON-RPC method j01FormProcessor
    and renders the content into the existing page dom using the javascript
    method j01RenderContent.

    The difference between this and the JSONRPCButton is that this button uses
    a JQuery delegation pattern. This means we don't apply the javascript event
    handler again on loading content conatining such a button. This is
    important if the form tag itself is contained in the dom and not a part of
    the loaded JSON-RPC content.

    If the edit form does not contain the form tag (partial jsonrpc loaded
    content in a large form), we need to use another pattern then the default
    .on('click', ..) handler. Otherwise we whould (re)apply everytime the edit
    button get loaded without to remove them from the form tag. This is because
    the form contains the delegated event handler if we sould use the
    .on('click',...) handler.

    This button uses a simple click handler which will get removed including
    it's click handler if the button get removed.

    IMPORTANT: If you use more then one JSONRPC form in one page you need to
    use different widget and button prefixes. Otherwise the button click
    handler get blocked if the same button is used in loaded open forms.
    The widget prefix is important if you use the same form field names in 2
    different forms loaded in the same page.

    The widget prefix will get changed by define a prefixWidgets property
    and the buttons prefix will get changed by define a prefixButtons property.
    See the improved updateWidgets and updateActions methods in JSONRPCForm

    Note: the rendered button is supported by p01.testbrowser testing!
    You can test this with something like:

        button = browser.getControl(label=None, name=None, index=None)
        button.click()

    """

    typ = 'JSONRPCClickButton'
    method = 'j01FormProcessor'
    onSuccess = 'j01RenderContentSuccess'
    onError = 'j01RenderContentError'

    def getInputEnterJavaScript(self, form, request):
        # not supported yet
        return ""

    def getJavaScript(self, action, request):
        # replace dots in id with '\\.' See jquery.com for details
        return """
            $('#%(action)s').click(function(){
                var $btn = $(this);
                if (!$btn.prop('disabled')) {%(trigger)s
                    var data = $('#%(form)s').j01FormToArray('%(handler)s');
                    proxy = getJSONRPCProxy('%(url)s');
                    proxy.addMethod(%(arguments)s);
                    proxy.%(method)s(data);
                }
                return false;
            });
            """ % {'form': action.form.id.replace('.', '\\\.'),
                   'action': action.id,
                   'handler': self.__name__,
                   'method': self.method,
                   'trigger': self.getTrigger(action),
                   'url': self.getURL(action.form, request),
                   'arguments': self.getMethodArguments(request),
                   }


# simple content loader button
@zope.interface.implementer(interfaces.IJSONRPCButton)
class JSONRPCContentButton(JSONRPCButton):
    """JSONRPC content loader button

    This button will load and render content via JSON-RPC from a JSONRPC page
    or form. The form does not get processed the form actions if we get the
    content from a form.

    The JSON-RPC method j01LoadContent get used for loading content.

    The callback method j01RenderContent is responsible for render the given
    content to the target defined by contentTargetExpression

    The optional urlGetter method can get used for define the url where the
    form content get loaded. By default the urlGetter uses the built in
    page url based on the pageURL property.

    This button requires the following javascript files:

    - jquery.js >= 1.7.0
    - j01.proxy.js
    - j01.jsonrpc.js

    NOTE: The form using this button must contain the form tag. If the form
    tag doesn't get rendered within each new request, the button click handler
    will get applied more then once to the form and doesn't work.

    Note: the rendered button is supported by p01.testbrowser testing!
    You can test this with something like:

        button = browser.getControl(label=None, name=None, index=None)
        button.click()

    """

    typ = 'JSONRPCContentButton'
    method = 'j01LoadContent'
    onSuccess = 'j01RenderContentSuccess'
    onError = 'j01RenderContentError'

    def getInputEnterJavaScript(self, form, request):
        """Returns the input enter JavaScript code."""
        # replace dotted id with '\\.' See jquery.com for details
        return """
            $('#%(form)s').on('keypress', 'input', function(e){
                if(!e){e = window.event;}
                key = e.which ? e.which : e.keyCode;
                if (key == 13) {
                    proxy = getJSONRPCProxy('%(url)s');
                    proxy.addMethod(%(arguments)s);
                    proxy.%(method)s();
                    return false;
                }
            });
            """ % {'form': form.id.replace('.', '\\\.'),
                   'url': self.getURL(form, request),
                   'method': self.method,
                   'arguments': self.getMethodArguments(request),
                   }

    def getJavaScript(self, action, request):
        """Returns the button javascript code"""
        # replace dotted id with '\\.' See jquery.com for details
        return """
            $('#%(form)s').on('click', '#%(action)s', function(){
                var $btn = $('#%(action)s');
                if (!$btn.prop('disabled')) {%(trigger)s
                    proxy = getJSONRPCProxy('%(url)s');
                    proxy.addMethod(%(arguments)s);
                    proxy.%(method)s();
                }
                return false;
            });
            """ % {'form': action.form.id.replace('.', '\\\.'),
                   'action': action.id,
                   'trigger': self.getTrigger(action),
                   'method': self.method,
                   'url': self.getURL(action.form, request),
                   'arguments': self.getMethodArguments(request),
                   }


# form processing button
@zope.interface.implementer(interfaces.IJSButton)
class CloseButton(TriggerMixin, JSButton):
    """Generic close button which will remove content from a html page.

    Note: the rendered button is supported by p01.testbrowser testing!
    You can test this with something like:

        button = browser.getControl(label=None, name=None, index=None)
        button.click()

    The expression defines a JQuery selector. The content contained will
    get removed from the dom on button click.

    This button requires the following javascript files:

    - jquery.js >= 1.7.0
    """

    typ = 'CloseButton'
    expression = 'form'
    animation = 'slideUp'
    duration = 400

    def __init__(self, *args, **kwargs):
        # apply expression
        if 'expression' in kwargs:
            self.expression = kwargs['expression']
            del kwargs['expression']
        # animation
        if 'animation' in kwargs:
            self.animation = kwargs['animation']
            del kwargs['animation']
        if self.animation not in ['slideUp', 'slideDown',]:
            raise ValueError("animation must be one of slideUp, slideDown")
        # apply duration
        if 'duration' in kwargs:
            self.duration = kwargs['duration']
            del kwargs['duration']
        super(CloseButton, self).__init__(*args, **kwargs)

    def getData(self, action, request):
        """Returns additional data attibutes and the testing control type

        NOTE: this button data attributes are used for testing. The Browser
        located in p01.testbrowser.browser is able to handle the button clicks
        using python testing hook methods.
        """
        data = super(CloseButton, self).getData(action, request)
        if request.get('paste.testing'):
            data['j01-testing-typ'] = self.typ
            data['j01-testing-expression'] = self.expression
        return data

    def getInputEnterJavaScript(self, form, request):
        # not supported yet
        return ""

    def getJavaScript(self, action, request):
        # replace dots in id with '\\.' See jquery.com for details
        return """
            $('#%(action)s').click(function(){
                var $btn = $(this);
                if (!$btn.prop('disabled')) {%(trigger)s
                    var part = $('%(expression)s');
                    part.%(animation)s(%(duration)s, function() {
                        part.empty();
                        part.show();
                    });
                }
                return false;
            });
            """ % {'form': action.form.id.replace('.', '\\\.'),
                   'action': action.id,
                   'trigger': self.getTrigger(action),
                   'expression': self.expression,
                   'animation': self.animation,
                   'duration': self.duration,
                   }


def canDelete(form):
    if hasattr(form, 'supportsDelete'):
        return form.supportsDelete
    return True


class IJSONRPCButtons(zope.interface.Interface):

    add = JSONRPCButton(
        title=_(u'Add')
        )

    applyChanges = JSONRPCButton(
        title=_(u'Apply')
        )

    delete = JSONRPCButton(
        title=_(u'Delete'),
        condition=canDelete
        )

    search = JSONRPCButton(
        title=_(u'Search')
        )

    cancel = JSONRPCButton(
        title=_(u'Cancel')
        )

    close = CloseButton(
        title=_(u'Close')
        )
